"""
Routines for easily creating spherical layers.
"""

# pylint: disable=E1129

import unittest

import numpy as np
import tensorflow as tf

from .layers import sphere_conv, ga_softmax
from .primitives import cos2d

class SphereConvTest(unittest.TestCase):
    """
    Tests for sphere_conv layers.
    """
    def test_cosines(self):
        """
        Test the cosine distances produced by the layer.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                images = tf.random_normal((3, 8, 9, 5), dtype=tf.float64)
                actual = sphere_conv(images,
                                     filters=4,
                                     kernel_size=(3, 2),
                                     strides=2,
                                     padding='same',
                                     variant='cosine',
                                     regularization=0.1)
                self.assertEqual(actual.dtype, tf.float64)
                kernels = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES,
                                            scope="sphere_conv/kernel")[0]
                expected = cos2d(images, kernels, [1, 2, 2, 1], 'SAME')
                sess.run(tf.global_variables_initializer())
                actual, expected = sess.run((actual, expected))
                self.assertTrue(np.allclose(expected, actual))

    def test_regularizer(self):
        """
        Basic regularizer tests.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                images = tf.random_normal((3, 8, 9, 5), dtype=tf.float64)
                sphere_conv(images, 4, 3, regularization=0.1)
                sphere_conv(images, 4, 3,
                            kernel_initializer=tf.contrib.layers.xavier_initializer(),
                            regularization=0.1)
                sess.run(tf.global_variables_initializer())
                orthog_reg, normal_reg = sess.run(tf.losses.get_regularization_losses())
                self.assertTrue(np.allclose(orthog_reg, 0))
                self.assertFalse(np.allclose(normal_reg, 0))

class GASoftmaxTest(unittest.TestCase):
    """
    Tests for the GA-Softmax.
    """
    def test_logits_1d(self):
        """
        Test logit values in a 1-dimensional case.
        """
        for variant in ['linear', 'cosine', 'sigmoid']:
            with tf.Graph().as_default():
                with tf.Session() as sess:
                    inputs = tf.constant([[2]], dtype=tf.float64)
                    logits = ga_softmax(inputs, 2, variant=variant, margin=3)
                    sess.run(tf.global_variables_initializer())
                    weights = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES,
                                                scope='ga_softmax/kernel')[0]
                    sess.run(tf.assign(weights, np.array([[1, -1]], dtype='float64')))
                    actual = sess.run(logits)
                    expected = np.array([2, -2], dtype='float64')
                    self.assertTrue(np.allclose(actual, expected))

    def test_margin_loss(self):
        """
        Test the loss term with a margin.
        """
        scewl = tf.nn.softmax_cross_entropy_with_logits
        for variant in ['linear', 'cosine', 'sigmoid']:
            for margin in [1.5, 2, 2.5, 3, 3.5, 4]:
                with tf.Graph().as_default():
                    with tf.Session() as sess:
                        inputs = tf.constant([[1, 2], [3, 4]], dtype=tf.float64)
                        labels = tf.constant([[0.1, 0.7, 0.2], [0, 0, 1]], dtype=tf.float64)
                        logits, losses = ga_softmax(inputs, 3,
                                                    labels=labels,
                                                    variant=variant,
                                                    margin=margin)
                        sess.run(tf.global_variables_initializer())
                        weights = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES,
                                                    scope='ga_softmax/kernel')[0]
                        sess.run(tf.assign(weights, np.array([[0.5, 0.7, 0.3], [1, 1, -1]],
                                                             dtype='float64')))
                        logits, losses = sess.run((logits, losses))
                        logit_losses = sess.run(scewl(labels=labels, logits=logits))
                        self.assertTrue((logit_losses < losses).all())



"""
Routines for easily creating spherical layers.
"""

from math import pi

import tensorflow as tf

from .primitives import cos1d, cos2d, sigmoid_nonlinearity

# pylint: disable=R0913
def sphere_conv(inputs,
                filters,
                kernel_size,
                strides=(1, 1),
                padding='valid',
                variant='linear',
                sigmoid_k=None,
                kernel_initializer=tf.orthogonal_initializer(),
                bias_initializer=tf.zeros_initializer(),
                regularization=None,
                name='sphere_conv'):
    """
    Create a SphereConv layer.

    This is similar to tf.layers.conv2d.

    Args:
      inputs: a 4-D input Tensor. Order is NHWC.
      filters: number of output filters.
      kernel_size: kernel dimensions (int or sequence).
      strides: convolutional stride (int or sequence).
      padding: 'valid' or 'same'.
      variant: 'linear', 'cosine', or 'sigmoid'.
      sigmoid_k: the `k` parameter for sigmoid layers.
        If None, a trainable variable is used.
        This value is broadcast as needed.
      kernel_initializer: initializer for the kernels.
      bias_initializer: initializer for the biases.
      regularization: the regularization coefficient.
        If not none, a regularization term is added to
        tf.GraphKeys.REGULARIZATION_LOSSES.
      name: name of the layer.
    """
    if not tf.contrib.framework.nest.is_sequence(kernel_size):
        kernel_size = (kernel_size, kernel_size)
    if not tf.contrib.framework.nest.is_sequence(strides):
        strides = (strides, strides)
    with tf.variable_scope(None, default_name=name):
        in_depth = int(inputs.get_shape()[-1])
        kernels = tf.get_variable('kernel',
                                  dtype=inputs.dtype,
                                  shape=(kernel_size[0], kernel_size[1], in_depth, filters),
                                  initializer=kernel_initializer)
        biases = tf.get_variable('bias',
                                 dtype=inputs.dtype,
                                 shape=(filters,),
                                 initializer=bias_initializer)
        if regularization:
            _add_kernel_regularizer(tf.reshape(kernels, (-1, filters)), regularization)
        cosines = cos2d(inputs, kernels, [1, strides[0], strides[1], 1], padding.upper())
        if variant == 'cosine':
            return biases + cosines
        elif variant == 'linear':
            return biases + 1 - (2/pi)*tf.acos(cosines)
        elif variant == 'sigmoid':
            sigmoid_k = _sigmoid_k_or_default(sigmoid_k, inputs.dtype)
            return biases + sigmoid_nonlinearity(tf.acos(cosines), sigmoid_k)
        else:
            raise ValueError('unknown variant: ' + variant)

# pylint: disable=R0914
def ga_softmax(inputs,
               outputs,
               labels=None,
               variant='linear',
               sigmoid_k=None,
               margin=1,
               kernel_initializer=tf.orthogonal_initializer(),
               name='ga_softmax'):
    """
    Create a generalized angular softmax layer.

    If `labels` is specified, a loss is computed and added
    to tf.GraphKeys.LOSSES.

    Args:
      inputs: a 2-D Tensor representing a batch of feature
        vectors to use for classification.
      outputs: number of output labels.
      labels: output labels for computing the loss.
        If specified, this is a batch of probability
        distributions.
      variant: 'linear', 'cosine', or 'sigmoid'.
      sigmoid_k: the `k` parameter for sigmoid layers.
        If None, a trainable variable is used.
        This value is broadcast as needed.
      margin: the margin coefficient. Values higher than 1
        encourage a large margin.
      kernel_initializer: initializer for the weights.
      name: name of the layer.

    Returns:
      If no labels are specified, a Tensor of logits.
      If labels are specified, a pair (logits, losses).
      The resulting logits do not depend on the margin.
    """
    with tf.variable_scope(None, default_name=name):
        weights = tf.get_variable('kernel',
                                  dtype=inputs.dtype,
                                  shape=(inputs.get_shape()[-1], outputs),
                                  initializer=kernel_initializer)
        angles = tf.acos(cos1d(inputs, weights))
        activation_fn = _ga_softmax_activation(variant, sigmoid_k, inputs.dtype)
        norms = tf.norm(inputs, axis=-1, keep_dims=True)
        logits = activation_fn(angles) * norms
        if labels is None:
            return logits
        margin_logits = activation_fn(angles * margin) * norms
        loss = None
        for i in range(outputs):
            sub_logits = tf.concat([logits[:, :i], margin_logits[:, i:i+1], logits[:, i+1:]],
                                   axis=-1)
            term = labels[:, i] * tf.nn.log_softmax(sub_logits)[:, i]
            if loss is None:
                loss = -term
            else:
                loss -= term
        tf.losses.add_loss(tf.reduce_mean(loss))
        return logits, loss

def _ga_softmax_activation(variant, sigmoid_k, dtype):
    """
    Get a function that applies the monotonically
    decreasing activation for a GA-Softmax.
    """
    if variant == 'linear':
        return lambda x: 1 - (2/pi)*x
    elif variant == 'cosine':
        return _repeated_cosine
    elif variant == 'sigmoid':
        sigmoid_k = _sigmoid_k_or_default(sigmoid_k, dtype)
        return lambda x: sigmoid_nonlinearity(x, sigmoid_k)
    else:
        raise ValueError('unknown variant: ' + variant)

def _add_kernel_regularizer(matrix, coeff):
    """
    Creates a regularization loss that encourages the
    columns of the matrix to be orthogonal.
    """
    dots = tf.matmul(tf.transpose(matrix), matrix)
    ident = tf.eye(int(matrix.get_shape()[-1]), dtype=matrix.dtype)
    diffs = tf.reduce_sum(tf.square(dots - ident))
    tf.losses.add_loss(diffs*coeff, loss_collection=tf.GraphKeys.REGULARIZATION_LOSSES)

def _sigmoid_k_or_default(sigmoid_k, dtype):
    """
    Return the sigmoid k constant or create a variable if
    the constant is None.
    """
    if sigmoid_k is not None:
        return sigmoid_k
    return tf.get_variable('k', dtype=dtype, initializer=tf.constant(0.5, dtype=dtype))

def _repeated_cosine(theta):
    """
    Monotonically decreasing piecewise cosine.
    """
    phase = tf.mod(theta, pi)
    offset = tf.floor(theta / pi)
    return tf.cos(phase) - 2*offset



"""
Spherical convolutional layers.
"""

from math import pi

import tensorflow as tf

def cos1d(inputs, kernels):
    """
    Compute the 1-D cosine distances between a batch of
    inputs and a batch of kernels.

    Args:
      inputs: a 2-D Tensor of shape [batch x in_size].
      kernels: a 2-D Tensor of shape [in_size x out_size].
      epsilon: small number to prevent divisions by zero.

    Returns:
      A 2-D Tensor of shape [batch x out_size] containing
        cosine distances between each input vector and all
        the kernels.
    """
    norm_inputs = inputs / _always_nonzero(tf.norm(inputs, axis=-1, keep_dims=True))
    norm_kernels = kernels / _always_nonzero(tf.norm(kernels, axis=0))
    return tf.matmul(norm_inputs, norm_kernels)

def cos2d(inputs, filters, strides, padding):
    """
    Compute the 2-D convolutional cosine distances between
    the filters and the input patches.

    This mimics tf.nn.conv2d.

    Args:
      inputs: a 4-D input Tensor.
      filters: a 4-D filter Tensor.
      strides: a 1-D stride Tensor of length 4.
      padding: a string, 'SAME', or 'VALID'.
      epsilon: small number to prevent divisions by zero.

    Returns:
      A Tensor with filter-patch cosine distances along
      the inner-most dimension.
    """
    filter_height, filter_width, _, _ = filters.get_shape()
    patches = tf.extract_image_patches(images=inputs,
                                       ksizes=[1, filter_height, filter_width, 1],
                                       strides=strides,
                                       rates=[1, 1, 1, 1],
                                       padding=padding)
    norm_patches = patches / _always_nonzero(tf.norm(patches, axis=-1, keep_dims=True))
    norm_filters = tf.reshape(filters, (-1, tf.shape(filters)[-1]))
    norm_filters /= _always_nonzero(tf.norm(norm_filters, axis=0, keep_dims=True))
    return tf.einsum('abcd,de->abce', norm_patches, norm_filters)

def sigmoid_nonlinearity(angles, sigmoid_k):
    """
    Compute a sigmoid SphereConv from angles.

    Args:
      angles: an N-D Tensor of angles.
      sigmoid_k: a Tensor of curvature coefficients.
        This is broadcast to match the shape of angles.
    """
    pi_coeff = -pi / (2 * sigmoid_k)
    scale = (1 + tf.exp(pi_coeff)) / (1 - tf.exp(pi_coeff))
    main_exp = tf.exp(angles/sigmoid_k + pi_coeff)
    return scale * (1 - main_exp) / (1 + main_exp)

def _always_nonzero(value, epsilon=1e-2):
    """
    Replace zero entries in value with a tiny epsilon.
    """
    return tf.where(tf.equal(value, 0), tf.zeros_like(value)+epsilon, value)



"""
Setup the package.
"""

from setuptools import setup

setup(
    name='spherenet',
    version='0.0.3',
    description='Deep Hyperspherical Learning.',
    long_description='Implementation of https://arxiv.org/abs/1711.03189v1.',
    url='https://github.com/unixpickle/spherenet',
    author='Alex Nichol',
    author_email='unixpickle@gmail.com',
    license='BSD',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='ai neural networks machine learning ml',
    packages=['spherenet'],
    install_requires=['numpy'],
    extras_require={
        "tf": ["tensorflow>=1.0.0"],
        "tf_gpu": ["tensorflow-gpu>=1.0.0"]
    }
)



"""
Tests for spherical convolutions.
"""

import unittest

import numpy as np
import tensorflow as tf

from .primitives import cos1d, cos2d

class Cos1DTest(unittest.TestCase):
    """
    Tests for the cos1d() primitive.
    """
    def test_asym(self):
        """
        Test diversely shaped inputs/kernels.
        """
        inputs = np.random.normal(size=(3, 5))
        filters = np.random.normal(size=(5, 2))
        expected = np.zeros((3, 2), dtype='float64')
        for batch_idx in range(3):
            for kernel_idx in range(2):
                expected[batch_idx][kernel_idx] = cos_dist(inputs[batch_idx],
                                                           filters[:, kernel_idx])
        actual = cos1d(tf.constant(inputs), tf.constant(filters))
        with tf.Session() as sess:
            self.assertTrue(np.allclose(expected, sess.run(actual)))

class Cos2DTest(unittest.TestCase):
    """
    Tests for the cos2d() primitive.
    """
    def test_single_patch(self):
        """
        Test when a single square patch fills up the
        entire input image.
        """
        image = np.random.normal(size=(3, 4, 4, 2))
        filters = np.random.normal(size=(4, 4, 2, 5))
        expected = np.zeros((3, 1, 1, 5), dtype='float64')
        for image_idx in range(image.shape[0]):
            for filter_idx in range(filters.shape[-1]):
                vec1 = image[image_idx, :, :, :]
                vec2 = filters[:, :, :, filter_idx]
                expected[image_idx, 0, 0, filter_idx] = cos_dist(vec1, vec2)
        actual = cos2d(tf.constant(image), tf.constant(filters), [1, 1, 1, 1], 'VALID')
        with tf.Session() as sess:
            self.assertTrue(np.allclose(expected, sess.run(actual)))

    def test_asym_slided(self):
        """
        Test weirdly shaped patches with strides.
        """
        image = np.random.normal(size=(5, 4, 5, 2))
        filters = np.random.normal(size=(2, 3, 2, 7))
        expected = np.zeros((5, 2, 2, 7), dtype='float64')
        for image_idx in range(expected.shape[0]):
            for filter_idx in range(expected.shape[3]):
                img = image[image_idx, :, :, :]
                filt = filters[:, :, :, filter_idx]
                expected[image_idx, 0, 0, filter_idx] = cos_dist(filt, img[0:2, 0:3])
                expected[image_idx, 1, 0, filter_idx] = cos_dist(filt, img[2:4, 0:3])
                expected[image_idx, 0, 1, filter_idx] = cos_dist(filt, img[0:2, 2:5])
                expected[image_idx, 1, 1, filter_idx] = cos_dist(filt, img[2:4, 2:5])
        actual = cos2d(tf.constant(image), tf.constant(filters), [1, 2, 2, 1], 'VALID')
        with tf.Session() as sess:
            self.assertTrue(np.allclose(expected, sess.run(actual)))

def cos_dist(vec1, vec2):
    """
    Compute the cosine distance between two tensors.
    """
    vec1 = vec1.flatten()
    vec2 = vec2.flatten()
    vec1 = vec1 / np.linalg.norm(vec1)
    vec2 = vec2 / np.linalg.norm(vec2)
    return np.dot(vec1, vec2)

if __name__ == '__main__':
    unittest.main()
